# coding=utf8
""" CLI Module

Holds methods for doing CLI things
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

class ProgressBar(object):
	"""Progress Bar

	Generates a progress bar for use in terminals
	"""

	def __init__(self, title="Progress Bar", total=100, sections=20, start=0):
		"""Constructor

		Initialises the instance of the object

		Arguments:
			title (str): The text before the progress bar
			total (int): The total count of items
			sections (int): The number of sections in the progress bar
			start (int): Optional, defaults to 0

		Returns:
			ProgressBar
		"""

		# If any values are invalid
		if not isinstance(total, int) or total < 1:
			raise ValueError('total must be at least 1 (one) or higher')
		if not isinstance(sections, int) or sections < 1:
			raise ValueError('sections must be at least 1 (one) or higher')
		if not isinstance(start, int) or start < 0:
			raise ValueError('start must be an unsigned integer')

		# Store the title and count
		self.title = title
		self.total = total
		self.sections = sections
		self.count = start

		# Figure out the steps from the sections
		self.steps = self.total / self.sections

		# Calculate percentage
		self.perc = round((self.count / self.total) * 100.0, 1)

		# Draw
		self.draw()

	def __add__(self, other):
		"""+ (overloaded)

		Handles adding to the count

		Arguments:
			other (uint): Should be an unsigned number

		Returns:
			None
		"""

		# Verify the other
		if not isinstance(other, int) or other < 1:
			raise ValueError('other must be 1 (one) or greater')

		# Add it to the count
		self.count += other

		# Recalculate the percentage
		self.perc = round((self.count / self.total) * 100.0, 1)

		# Update the drawing
		self.draw()

		# If we reached maximum
		if self.count >= self.total:
			print('')

	def clear(self):
		"""Clear

		Clears the screen

		Returns:
			None
		"""
		print('')

	def draw(self):
		"""Draw

		Draws the progress bar in the terminal

		Returns:
			None
		"""

		# Calculate the steps done
		iDone = int(self.count // self.steps)
		iDiff = self.count % self.steps

		# If the diff is more than half a step, add a step
		if self.steps - iDiff <= iDiff:
			iDone += 1

		# Generate the actual progress bar
		sDiff = ('\u2588' * iDone) + (' ' * (self.sections - iDone))

		# Print everything
		print("\r%s [%s] %.1f%%" % (self.title, sDiff, self.perc), end=' ')



# coding=utf8
"""Image

Module to manipulate images and photos, requires the Python Pillow library
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2016-11-11"

# Ouroboros imports
from tools import crop as crop_tool, fit as fit_tool

# Python imports
import re
from io import BytesIO

# Pip Imports
import piexif
from PIL import Image as Pillow, ImageFile as PillowFile
PillowFile.LOAD_TRUNCATED_IMAGES = True

# exif rotation tag
ORIENTATION_TAG = 0x0112

# Regex for valid dimension strings
DIMENSIONS_REGEX = re.compile(r'^(?:[1-9]\d+)?x(?:[1-9]\d+)?$')

# Rotation sequences based on exif orientation flag
SEQUENCES = [
	[],
	[Pillow.FLIP_LEFT_RIGHT],
	[Pillow.ROTATE_180],
	[Pillow.FLIP_TOP_BOTTOM],
	[Pillow.FLIP_LEFT_RIGHT, Pillow.ROTATE_90],
	[Pillow.ROTATE_270],
	[Pillow.FLIP_TOP_BOTTOM, Pillow.ROTATE_90],
	[Pillow.ROTATE_90]
]

def apply_rotation(image):
	"""Apply Rotation

	Uses exif data to rotate the image to the proper position, will lose exif
	data and might lose quality

	Arguments:
		image (str): A raw image as a string

	Returns:
		str
	"""

	# Load the image into a new BytesIO
	sImg = BytesIO(image)
	sNewImg = BytesIO(b'')

	# Create a new Pillow instance from the raw data
	oImg = Pillow.open(sImg)

	# Store the image format
	sFormat = oImg.format

	# Get the proper sequence
	try:
		lSeq = SEQUENCES[oImg._getexif()[ORIENTATION_TAG] - 1]

		# Transpose the image
		for i in lSeq:
			oImg = oImg.transpose(i)

		# Save the image using the same format as we got it in
		oImg.save(sNewImg, sFormat)

		# Get the raw bytes
		sRet = sNewImg.getvalue()

	# If there's no sequence, return the image as is
	except Exception as e:
		sRet = image

	# Cleanup
	oImg.close()
	sImg.close()
	sNewImg.close()

	# Return
	return sRet

def convert_to_jpeg(image, quality=90):
	"""Convert To JPEG

	Changes any valid image into a JPEG, loses exif data

	Arguments:
		image (str): A raw image as a string
		quality (uint): The quality, from 0 to 100, to save the image in

	Returns:
		str
	"""

	# Load the image into a new BytesIO
	sImg = BytesIO(image)

	# Create an empty BytesIO for the new image
	sNewImg = BytesIO(b'')

	# Create a new Pillow instance from the raw data
	oImg = Pillow.open(sImg)

	# If the mode is not valid
	if oImg.mode not in ('1','L','RGB','RGBA'):
		oImg = oImg.convert('RGB')

	# Save the new image as a JPEG
	oImg.save(sNewImg, 'JPEG', quality=quality, subsampling=0)

	# Pull out the raw string
	sRet = sNewImg.getvalue()

	# Close the image
	oImg.close()

	# Return the new image
	return sRet

def info(image):
	"""Info

	Returns information about an image: resolution, length, type, and mime

	Arguments:
		image (str): A raw image as a string

	Returns:
		dict
	"""

	# Load the image into a new BytesIO
	sImg = BytesIO(image)

	# Create a new Pillow instance from the raw data
	oImg = Pillow.open(sImg)

	# Get the details
	dInfo = {
		"length": len(image),
		"mime": oImg.format in Pillow.MIME and Pillow.MIME[oImg.format] or None,
		"resolution": {
			"width": oImg.size[0],
			"height": oImg.size[1]
		},
		"type": oImg.format
	}

	# Check for exif rotation data
	try:
		dInfo['orientation'] = oImg._getexif()[ORIENTATION_TAG]
	except Exception:
		dInfo['orientation'] = False

	# Check for exif data
	try:
		dInfo['exif'] = piexif.load(oImg.info['exif'])
	except Exception:
		dInfo['exif'] = None

	# Cleanup
	sImg.close()
	oImg.close()

	# Return the info
	return dInfo

def resize(image, dims, crop = False, quality = 90, format = None):
	"""Resize

	Given raw data and a size, a new image is created and returned as raw data

	Arguments:
		image (str): Raw image data to be loaded and resized
		dims (str|dict): New dimensions of the image, "WWWxHHH" or {"w":, "h":}
		crop (bool): Set to true to crop the photo rather than add whitespace
		quality (uint): The quality, from 0 to 100, to save the thumbnail in
		format (str): The optional format to use instead of the original

	Returns:
		str
	"""

	# Check the dimensions
	if not isinstance(dims, dict):
		if isinstance(dims, str):
			l = [i for i in dims.split('x')]
			dims = {"w": l[0], "h": l[1]}
		else:
			raise ValueError('dims')

	# Load the image into a new BytesIO
	sImg = BytesIO(image)
	sNewImg = BytesIO(b'')

	# Create a new Pillow instance from the raw data
	oImg = Pillow.open(sImg)

	# Store the format
	sFormat = format or oImg.format

	# Make sure the values are ints
	dims['w'] = int(dims['w'])
	dims['h'] = int(dims['h'])

	# Create a new blank image
	oNewImg = Pillow.new(oImg.mode, [dims['w'],dims['h']], (255,255,255,255))

	# If the image has an orientation
	try:
		lSeq = SEQUENCES[oImg._getexif()[ORIENTATION_TAG] - 1]

		# Transpose the image
		for i in lSeq:
			oImg = oImg.transpose(i)
	except Exception:
		pass

	# If we are cropping
	if crop:
		dResize = crop_tool(oImg.width, oImg.height, dims['w'], dims['h'])

	# Else, we are fitting
	else:
		dResize = fit_tool(oImg.width, oImg.height, dims['w'], dims['h'])

	# Resize the image
	oImg.thumbnail([dResize['w'], dResize['h']], Pillow.ANTIALIAS)

	# Get the offsets
	lOffset = ((dims['w'] - dResize['w']) // 2, (dims['h'] - dResize['h']) // 2)

	# Paste the resized image onto the new canvas
	oNewImg.paste(oImg, lOffset)

	# Save the new image to a BytesIO
	oNewImg.save(sNewImg, sFormat, quality=90, subsampling=0)

	# Pull out the raw string
	sReturn = sNewImg.getvalue()

	# Cleanup
	oNewImg.close()
	oImg.close()
	sImg.close()

	# Return the new string
	return sReturn



# coding=utf8
""" Date & Time Helper Module

Several useful helper methods for use with dates and times
"""

# Compatibility
from past.builtins import basestring

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2021-05-01"

# Python imports
from math import floor

# Pip imports
import arrow

def _to_arrow(val):
	"""To Arrow

	Converts a value to an Arrow instance for easier use

	Arguments:
		val (mixed): An value that can hopefully be turned into some sort of
						date/time

	Returns:
		arrow.arrow.Arrow
	"""

	# If we got a timestamp
	if isinstance(val, int):
		return arrow.get(val)

	# If we got a string
	if isinstance(val, basestring):

		# If it's only ten characters, add the time and timezone
		if len(val) == 10:
			return arrow.get('%sT00:00:00+00:00' % val)

		# If it's 19 characters, replace any space with T and add the timezone
		if len(val) == 19:
			return arrow.get('%+00:00' % val.replace(' ', 'T'))

		# If it's 25 characters, assume it's good
		if len(val) == 25:
			return arrow.get(val)

		# Raise an exception
		raise ValueError('Invalid date string', val)

	# If it's already an arrow instance
	if isinstance(val, arrow.arrow.Arrow):
		return val

	# Raise an exception
	raise ValueError('Invalid date/time', val)

def age(dob):
	"""Age

	Returns the current age of someone based on today's date and their DOB.
	This method is not %100 accurate, but it's good enough for 99% of cases

	Arguments:
		dob (Arrow|uint|str): The date of birth of the person

	Returns:
		uint
	"""

	# Make sure we have an arrow instance
	oDOB = _to_arrow(dob)

	# Get the delta from today
	oDelta = arrow.get() - oDOB

	# Return the age
	return floor(oDelta.days / 365.25)

def date(d):
	"""Date

	Returns a string in YYYY-MM-DD format from a timestamp, date string, or
	arrow instance

	Arguments:
		d (Arrow|uint|str): The date to format

	Returns:
		str
	"""

	# Make sure we have an arrow instance
	oD = _to_arrow(d)

	# Return the date string
	return oD.format('YYYY-MM-DD')

def date_increment(days=1, from_=None):
	"""Date Increment

	Returns a date incremented by the given days. Use negative to decrement.

	Arguments:
		days (int): The number of days to increment (or decrement) by
		from_ (mixed): Optional, the date to increment from, else today

	Returns:
		arrow.arrow.Arrow
	"""

	# If we got a from
	oDate = from_ and _to_arrow(from_) or arrow.get()

	# Increment the date and return it
	return oDate.shift(days=days)

def datetime(d):
	"""Date Time

	Returns a string in YYYY-MM-DD HH:mm:ss format from a timestamp, date
	string, or arrow instance

	Arguments:
		d (Arrow|uint|str): The date to format

	Returns:
		str
	"""

	# Make sure we have an arrow instance
	oD = _to_arrow(d)

	# Return the date/time string
	return oD.format('YYYY-MM-DD HH:mm:ss')

def time_elapsed(seconds, opts=None):
	"""Time Elapsed

	Returns seconds in a human readable format with several options to show/hide
	hours/minutes/seconds

	Arguments:
		seconds (uint): The seconds to convert to ((HH:)mm:)ss
		opts (dict): Optional flags:
						show_minutes: default True
						show_seconds: default True
						show_zero_hours: default False
						show_zero_minutes: default False

	Returns:
		str
	"""

	# Get the hours and remaining seconds
	h, r = divmod(seconds, 3600)

	# Get the minutes and seconds
	m, s = divmod(r, 60)

	# Generate the flags
	bShowMinutes = not opts or 'show_minutes' not in opts or opts['show_minutes']
	bShowSeconds = not opts or 'show_seconds' not in opts or opts['show_seconds']
	bShowZeroHours = opts and 'show_zero_hours' in opts and opts['show_zero_hours']
	bShowZeroMinutes = opts and 'show_zero_minutes' in opts and opts['show_zero_minutes']

	# Init the list we'll turn into time
	lTime = None

	# If we have hours
	if h:

		# Start by adding hours
		lTime = [str(h)]

		# If we want to show minutes
		if bShowMinutes:
			lTime.append(m < 10 and ('0%d' % m) or str(m))

			# If we want to show seconds (can't show seconds if no minutes)
			if bShowSeconds:
				lTime.append(s < 10 and ('0%d' % s) or str(s))

	# Else, if we have minutes
	elif m:

		# Init the time
		lTime = []

		# If we want to show zero hours
		if bShowZeroHours:
			lTime.append('0')

		# If we want to show minutes
		if bShowMinutes:
			lTime.append((bShowZeroHours and m < 10) and ('0%d' % m) or str(m))

			# If we want to show seconds (can't show seconds if no minutes)
			if bShowSeconds:
				lTime.append(s < 10 and ('0%d' % s) or str(s))

	# Else, we only have seconds
	else:

		# Init the time
		lTime = []

		# If we want to show zero hours
		if bShowZeroHours:
			lTime.extend(['0', '00'])

		# Else, if we want to show zero minutes
		elif bShowZeroMinutes:
			lTime.append('0')

		# If we want to show seconds
		if bShowMinutes and bShowSeconds:
			lTime.append(
				((bShowZeroMinutes or bShowZeroHours) and s < 10) and \
				('0%d' % s) or \
				str(s)
			)

	# Put them all together and return
	return ':'.join(lTime)



# coding=utf8
""" Services Module

Tools to create and communicate with Micro Services
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Ouroboros imports
import jsonb
import undefined

# Python imports
from datetime import datetime
from hashlib import sha1
import re
from time import sleep, time

# Pip imports
import requests

# Module imports
from . import Errors

__mbVerbose = False
"""Verbose Flag"""

__mdRegistered = {}
"""Registered Services"""

__miInternal = 0
"""Internal Key timeout in seconds"""

__msSalt = None
"""Internal Key Salt"""

__funcToRequest = {
	'create': [requests.post, 'POST'],
	'delete': [requests.delete, 'DELETE'],
	'read': [requests.get, 'GET'],
	'update': [requests.put, 'PUT']
}
"""Map functions to REST types"""

def request(service, action, path, req = {}):
	"""Request

	Method to convert REST requests into HTTP requests

	Arguments:
		service (str): The service we are requesting data from
		action (str): The action to take on the service
		path (str): The path of the request
		req (dict): The request details: 'data', 'session', and 'enviroment'

	Raises:
		ServiceException

	Return:
		Response
	"""

	# If we have a registered service
	if service in __mdRegistered:

		# If the service is locally loaded
		if 'instance' in __mdRegistered[service]:

			# If verbose requested
			if __mbVerbose:
				print('%s: Calling %s.%s("%s", %s)' % (
					str(datetime.now()),
					service,
					action,
					path,
					'data' in req and jsonb.encode(req['data'], 2) or 'NONE')
				)

			# Directly call the action
			oResponse = getattr(__mdRegistered[service]['instance'], action)(
				path, req
			)

			# If verbose requested
			if __mbVerbose:
				print('%s: Returning %s\n' % (
					str(datetime.now()),
					jsonb.encode(oResponse.to_dict(), 2))
				)

		# Else if the service is running elsewhere
		else:

			try: __funcToRequest[action]
			except KeyError: Error(Errors.SERVICE_ACTION, action)

			# Generate the URL to reach the service
			sURL = __mdRegistered[service]['url'] + path

			# Init the data and headers
			sData = ''
			dHeaders = {
				'Content-Length': '0',
				'Content-Type': 'application/json; charset=utf-8'
			}

			# If the data was passed
			if 'data' in req and req['data']:

				# Convert the data to JSON and store the length
				sData = jsonb.encode(req['data'])
				dHeaders['Content-Length'] = str(len(sData))

			# If we have a session, add the ID to the headers
			if 'session' in req and req['session']:
				dHeaders['Authorization'] = req['session'].id()

			# Try to make the request and store the response
			iAttempts = 0
			while True:
				iAttempts += 1
				try:
					oRes = __funcToRequest[action][0](sURL, data=sData, headers=dHeaders)

					# If the request wasn't successful
					if oRes.status_code != 200:

						# If we got a 401
						if oRes.status_code == 401:
							return Response.from_json(oRes.content)
						else:
							return Error(Errors.SERVICE_STATUS, '%d: %s' % (oRes.status_code, oRes.content))

					# If we got the wrong content type
					if oRes.headers['Content-Type'].lower() != 'application/json; charset=utf-8':
						return Error(Errors.SERVICE_CONTENT_TYPE, '%s' % oRes.headers['content-type'])

					# Success, break out of the loop
					break

				# If we couldn't connect to the service
				except requests.ConnectionError as e:

					# If we haven't exhausted attempts
					if iAttempts < 3:

						# Wait for a second
						sleep(1)

						# Loop back around
						continue

					# We've tried enough, return an error
					return Error(Errors.SERVICE_UNREACHABLE, str(e))

			# Else turn the content into an Response and return it
			oResponse = Response.from_json(oRes.text)

		# Return the Response of the request
		return oResponse

	# Service not registered
	else:
		raise ResponseException(error=(Errors.SERVICE_NOT_REGISTERED, service))

def create(service, path, req = {}):
	"""Create

	Make a POST request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data',
					'environment', and 'session'

	Returns:
		Response
	"""
	return request(service, 'create', path, req)

def delete(service, path, req = {}):
	"""Delete

	Make a DELETE request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data',
					'environment', and 'session'

	Returns:
		Response
	"""
	return request(service, 'delete', path, req)

def internal_key(key = None):
	"""Internal Key

	Generates or validates an internal key so services can communicate with
	each other

	Arguments:
		key (str): Passed to validate a key

	Returns:
		bool
	"""

	# Pull in salt
	global __msSalt

	# Generate a timestamp
	iTime = int(time())

	# If no key was passed
	if key is None:

		# Generate a timestamp and store it as a string
		sTime = str(iTime)

		# Generate a sha1 from the salt and parts of the time
		sSHA1 = sha1(sTime[5:].encode('utf-8') + __msSalt.encode('utf-8') + sTime[:5].encode('utf-8')).hexdigest()

		# Generate a key using the sha1 and the time
		return sSHA1 + ':' + sTime

	# If the key was passed
	else:
		try:
			# Split the key into sha1 and timestamp
			sSHA1_, sTime = key.split(':')

			# If the time is not close enough
			if iTime - int(sTime) > __miInternal:
				return False

			# Generate a sha1 from the salt and parts of the time
			sSHA1 = sha1(sTime[5:].encode('utf-8') + __msSalt.encode('utf-8') + sTime[:5].encode('utf-8')).hexdigest()

			# If the sha1s match return true
			return sSHA1 == sSHA1_

		# If something went wrong, return false
		except Exception:
			return False

def read(service, path, req = {}):
	"""Read

	Make a GET request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data',
					'environment', and 'session'

	Returns:
		Response
	"""
	return request(service, 'read', path, req)

def register(services, restconf, salt, internal=5):
	"""Register

	Takes a dictionary of services to their instances, or None for remote
	services which will be found via the config

	Arguments:
		services (dict): Services being registered
		restconf (dict): Configuration variables for remote services
		salt (str): The salt used for internal key generation
		internal (uint): The time in seconds an internal key is valid

	Raises:
		ValueError

	Returns:
		None
	"""

	# Pull in the global salt and internal variables and set them
	global __msSalt, __miInternal
	__msSalt = salt
	__miInternal = internal

	# If we didn't get a dictionary
	if not isinstance(services, dict):
		raise ValueError('services')

	# Loop through the list of services to register
	for k,v in services.items():

		# If verbose requested
		if __mbVerbose: print('Registering service "%s": ' % str(k), end='')

		# If we received a local instance
		if isinstance(v, Service):

			# Store it
			__mdRegistered[k] = {"instance":v}

			# If verbose requested
			if __mbVerbose:	print('instance')

			# Call the services initialise method
			v.initialise()

		# Else the service is remote
		elif v is None:

			# Make sure we have the service
			if k not in restconf:
				raise ValueError('services.%s' % k)

			# Store it
			__mdRegistered[k] = {"url":restconf[k]['url']}

			# If verbose mode is on
			if __mbVerbose:	print('%s' % __mdRegistered[k]['url'])

		# Else, the value is invalid
		else:
			raise ValueError('services.%s' % str(k))

def update(service, path, req = {}):
	"""Update

	Make a PUT request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data',
					'environment', and 'session'

	Returns:
		Response
	"""
	return request(service, 'update', path, req)

def verbose(flag=True):
	"""Verbose

	Puts Services in verbose mode for easy tracking of requests

	Arguments:
		flag (bool): defaults to True

	Returns:
		None
	"""

	global __mbVerbose

	if __mbVerbose and not flag:
		print('Service verbose mode will be turned off')

	__mbVerbose = flag

	if __mbVerbose:
		print('Service verbose mode has been turned on')

class Response(object):
	"""Response

	Represents a standard result from any/all requests
	"""

	def __init__(self, data = undefined, error = undefined, warning = undefined):
		"""Constructor

		Initialises a new Response instance

		Arguments:
			data (mixed): If a request returns data this should be set
			error (mixed): If a request has an error, this can be filled with
				a code and message string
			warning (mixed): If a request returns a warning this should be set

		Raises:
			ValueError

		Returns:
			Response
		"""

		# If there's data, store it as is
		if data is not undefined:
			self.data = data

		# If there's an error, figure out what type
		if error is not undefined:

			# If we got an int, it's a code with no message string
			if isinstance(error, int):
				self.error = {'code': error, 'msg': ''}

			# If we got a string, it's a message with no code
			elif isinstance(error, str):
				self.error = {'code': 0, 'msg': error}

			# If it's a tuple, 0 is a code, 1 is a message
			elif isinstance(error, tuple):
				self.error = {'code': error[0], 'msg': error[1]}

			# If we got a dictionary, assume it's already right
			elif isinstance(error, dict):
				self.error = error

			# If we got an exception
			elif isinstance(error, Exception):

				# If we got another Response in the Exception, store the error
				#	from it
				if isinstance(error.args[0], Response):
					self.error = error.args[0].error

				# Else, try to pull out the code and message
				else:
					self.error = {'code': error.args[0], 'msg': ''}
					if len(error.args) > 1: self.error['msg'] = error.args[1]

			# Else, we got something invalid
			else:
				raise ValueError('error')

		# If there's a warning, store it as is
		if not warning is undefined:
			self.warning = warning

	def __str__(self):
		"""str

		Python magic method to return a string from the instance

		Returns:
			str
		"""

		# Create a temp dict
		dRet = {}

		# If there's data
		try: dRet['data'] = self.data
		except AttributeError: pass

		# If there's an error
		try: dRet['error'] = self.error
		except AttributeError: pass

		# If there's a warning
		try: dRet['warning'] = self.warning
		except AttributeError: pass

		# Convert the dict to JSON and return it
		return jsonb.encode(dRet)

	def data_exists(self):
		"""Data Exists

		Returns True if there is data in the Response

		Returns:
			bool
		"""
		try: return self.data != None
		except AttributeError: return False

	def error_exists(self):
		"""Error Exists

		Returns True if there is an error in the Response

		Returns:
			bool
		"""
		try: return self.error != None
		except AttributeError: return False

	@classmethod
	def from_dict(cls, val):
		"""From Dict

		Converts a dict back into an Response

		Arguments:
			val (dict): A valid dict

		Returns:
			Response
		"""

		# Create a new instance
		o = cls()

		# If there's data
		try: o.data = val['data']
		except KeyError: pass

		# If there's an error
		try: o.error = val['error']
		except KeyError: pass

		# If there's a warning
		try: o.warning = val['warning']
		except KeyError: pass

		# Return the instance
		return o

	@classmethod
	def from_json(cls, val):
		"""From JSON

		Tries to convert a string made from str() back into an Response

		Arguments:
			val (str): A valid JSON string

		Returns:
			Response
		"""

		# Try to convert the string to a dict
		try: d = jsonb.decode(val)
		except ValueError as e: raise ValueError('val', str(e))
		except TypeError as e: raise ValueError('val', str(e))

		# Return the fromDict result
		return cls.from_dict(d)

	def to_dict(self):
		"""To Dict

		Converts the Response into a dict

		Returns:
			dict
		"""

		# Init the return
		dRet = {}

		# Look for a data attribute
		try: dRet['data'] = self.data
		except AttributeError: pass

		# Look for an error attribute
		try: dRet['error'] = self.error
		except AttributeError: pass

		# Look for a warning attribute
		try: dRet['warning'] = self.warning
		except AttributeError: pass

		# Return the dict
		return dRet

	def warning_exists(self):
		"""Warning Exists

		Returns True if there is a warning in the Response

	Returns:
			bool
		"""
		try: return self.warning != None
		except AttributeError: return False

class Error(Response):
	"""Error

	Shorthand form of Response(error=)
	"""

	def __init__(self, code, msg = None):
		"""Constructor

		Initialises a new Response instance

		Arguments:
			code (uint): The error code
			msg (mixed): Optional message for more info on the error

		Returns:
			Error
		"""

		# Set the error code
		self.error = {
			'code': code,
			'msg': msg
		}

class ResponseException(Exception):
	"""Response Exception

	Stupid python won't let you raise anything that doesn't extend BaseException
	"""

	def __init__(self, data = undefined, error = undefined, warning = undefined):
		"""Constructor

		Dumb dumb python

		Arguments:
			data (mixed): If a request returns data this should be set
			error (mixed): If a request has an error, this can be filled with
				a code and message string
			warning (mixed): If a request returns a warning this should be set

		Returns:
			ResponseException
		"""

		# If we got a Response object
		if isinstance(data, Response):
			super().__init__(data)

		# Else, construct the Response and pass it to the parent
		else:
			super().__init__(Response(data, error, warning))

class Service(object):
	"""Service

	The object to build all Services from
	"""

	__pathToDef = {}
	"""Map of paths to function name"""

	__keyToError = {
		'data': Errors.SERVICE_NO_DATA,
		'session': Errors.SERVICE_NO_SESSION
	}
	"""Maps missing data keys to error codes"""

	__nounRegex = re.compile(r'([a-z]+(?:_[a-z]+)*)_(create|delete|read|update)')
	"""Regular Expression to match to valid service noun method"""

	def create(self, path, req):
		"""Create

		Create a new object

		Arguments:
			path (str): The path passed to the request
			req (dict): The request details, which can include 'data',
						'environment', and 'session'

		Return:
			Response
		"""

		# Generate the method name from the URI
		sMethod = self.path_to_method(path, '_create')

		# Try to find the method
		try:
			f = getattr(self, sMethod)

		# Method doesn't exist, URI is invalid
		except AttributeError as e:

			# If the method wasn't found
			if "'%s'" % sMethod in e.args[0]:
				return Response(error=(Errors.SERVICE_NO_SUCH_NOUN, 'POST %s' % path))
			else:
				raise

		# Try to call the method
		try:
			return f(req)

		# If we got a KeyError
		except KeyError as e:
			if e.args[0] in self.__keyToError:
				return Error(self.__keyToError[e.args[0]])
			raise e

		# Response thrown
		except ResponseException as e:
			return e.args[0]

	def delete(self, path, req):
		"""Delete

		Delete an existing object

		Arguments:
			path (str): The path passed to the request
			req (dict): The request details, which can include 'data',
						'environment', and 'session'

		Return:
			Response
		"""

		# Generate the method name from the URI
		sMethod = self.path_to_method(path, '_delete')

		# Try to find the method
		try:
			f = getattr(self, sMethod)

		# Method doesn't exist, URI is invalid
		except AttributeError as e:

			# If the method wasn't found
			if "'%s'" % sMethod in e.args[0]:
				return Response(error=(Errors.SERVICE_NO_SUCH_NOUN, 'DELETE %s' % path))
			else:
				raise

		# Try to call the method
		try:
			return f(req)

		# If we got a KeyError
		except KeyError as e:
			if e.args[0] in self.__keyToError:
				return Error(self.__keyToError[e.args[0]])
			raise e

		# Response thrown
		except ResponseException as e:
			return e.args[0]

	def initialise(self):
		"""Initialise

		Initialises the instance and returns itself for chaining

		Returns:
			Service
		"""

		# Return self for chaining
		return self

	def generateURLS(self, prefix = None):
		"""Generate URLS

		Finds all the appropriate functions on the Child service and returns
		a dictionary of URLs to bound functions

		Arguments:
			prefix (str): The optional prefix for urls

		Returns:
			{ str : bound method }
		"""

		# Fetch all available methods
		lURLs = []

		# Go through each available method
		for s in dir(self):

			# Try matching the proper format
			oMatch = self.__nounRegex.match(s)

			# If we have a match
			if oMatch:
				lURL = prefix and [prefix] or []
				lURL.extend(oMatch.group(1).split('_'))
				lURLs.append(['/'.join(lURL), oMatch.group(2)])

		# Returns the generated urls
		return lURLs

	@classmethod
	def install(cls):
		"""Install

		Installs any necessary DBs, configs, etc, needed by the Service when it
		is first installed

		Raises:
			NotImplementedError

		Returns:
			bool
		"""
		raise NotImplementedError('Must implement the "install" method')

	def read(self, path, req):
		"""Read

		Read an existing object

		Arguments:
			path (str): The path passed to the request
			req (dict): The request details, which can include 'data',
						'environment', and 'session'

		Return:
			Response
		"""

		# Generate the method name from the URI
		sMethod = self.path_to_method(path, '_read')

		# Try to find the method
		try:
			f = getattr(self, sMethod)

		# Method doesn't exist, URI is invalid
		except AttributeError as e:

			# If the method wasn't found
			if "'%s'" % sMethod in e.args[0]:
				return Response(error=(Errors.SERVICE_NO_SUCH_NOUN, 'GET %s' % path))
			else:
				raise

		# Try to call the method
		try:
			return f(req)

		# If we got a KeyError
		except KeyError as e:
			if e.args[0] in self.__keyToError:
				return Error(self.__keyToError[e.args[0]])
			raise e

		# Response thrown
		except ResponseException as e:
			return e.args[0]

	def update(self, path, req):
		"""Update

		Update an existing object

		Arguments:
			path (str): The path passed to the request
			req (dict): The request details, which can include 'data',
						'environment', and 'session'

		Return:
			Response
		"""

		# Generate the method name from the URI
		sMethod = self.path_to_method(path, '_update')

		# Try to find the method
		try:
			f = getattr(self, sMethod)

		# Method doesn't exist, URI is invalid
		except AttributeError as e:

			# If the method wasn't found
			if "'%s'" % sMethod in e.args[0]:
				return Response(error=(Errors.SERVICE_NO_SUCH_NOUN, 'PUT %s' % path))
			else:
				raise

		# Try to call the method
		try:
			return f(req)

		# If we got a KeyError
		except KeyError as e:
			if e.args[0] in self.__keyToError:
				return Error(self.__keyToError[e.args[0]])
			raise e

		# Response thrown
		except ResponseException as e:
			return e.args[0]

	@classmethod
	def path_to_method(cls, path, append=''):
		"""Path to Method

		Takes a path and converts it to the standard naming for Service methods

		Arguments:
			path (str): The path to parse
			append (str): If set, appended to method name

		Returns:
			str
		"""

		# If we already have the function name associated
		try:
			sRet = cls.__pathToDef[path]

		# Else, if we haven't generated it yet
		except KeyError:

			# Generate the name from the path
			sRet = ''
			iLen = len(path)
			i = 0
			while i < iLen:
				sRet += (path[i] in ['/', '-']) and '_' or path[i]
				i += 1

			# Store it for next time
			cls.__pathToDef[path] = sRet

		# Return the generated function name plus the append
		return sRet + append



# coding=utf8
"""SMTP

Wrapper for python smtp module
"""

# Compatibility
from past.builtins import basestring

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-17"

# Python imports
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate
import smtplib
import socket
from os.path import basename

# Init the local variables
__mdSMTP = {
	"host": "localhost",
	"port": 25,
	"tls": False,
	"passwd": ""
}

# Init the last error message
__msError = ''

# Create the defines
OK = 0
ERROR_UNKNOWN = -1
ERROR_CONNECT = -2
ERROR_LOGIN = -3

def _addresses(l):
	"""Addresses

	Takes a string or list of strings and returns them formatted for to:, cc:,
	or bcc:

	Arguments:
		l (str|str[]): The address or list of addresses

	Returns:
		str
	"""

	# If we got a list, tuple, or set
	if isinstance(l, (list,tuple,set)):
		return ', '.join(l)
	else:
		return l

def _to(l):
	"""To

	Converts all addresses passed, whether strings or lists, into one singular
	list

	Arguments:
		l (list(str|str[])): The list of addresses or lists of addresses

	Returns:
		list
	"""

	# Init the return list
	lRet = []

	# Go through each item in the list
	for m in l:

		# If we got a list
		if isinstance(m, (list,tuple)):
			lRet.extend(m)

		# Else, we got one address
		else:
			lRet.append(m)

	# Return the full list
	return lRet

def init(host="localhost", port=25, tls=False, user=None, passwd=None):
	"""init

	Called to change/set any SMTP information before sending out any e-mails

	Args:
		host (str): The hostname of the SMTP server
		port (uint): The port on the host
		tls (bool): Set to True if we need TLS
		user (str): The authorization username
		passwd (str): The authorization password

	Returns:
		None
	"""

	# Import the module var
	global __mdSMTP

	# If the host is set
	if host:
		__mdSMTP['host'] = host

	# If the port is set
	if port:
		__mdSMTP['port'] = port

	# If we need TLS
	if tls:
		__mdSMTP['tls'] = tls

	# If the user is set
	if user:
		__mdSMTP['user'] = user

	# If the passwd is set
	if passwd:
		__mdSMTP['passwd'] = passwd

def last_error():
	"""Last Error

	Returns the last error message if there is one

	Returns:
		str
	"""
	global __msError
	return __msError

def send(to, subject, text_body = None, html_body = None, from_='root@localhost', bcc=None, attachments=None):
	"""Send

	Sends an e-mail to one or many addresses

	Arguments:
		to (str|str[]): One or email addresses to send to
		subject (str): The email's subject
		text_body (str): The text version of the main content of the email
		html_body (str): The html version of the main content of the email
		from_ (str): The from address of the email, optional
		bcc (str|str[]): Blind carbon copy addresses, optional
		attachments

	Returns:
		bool
	"""

	# Opts
	dOpts = {}

	# Add available options
	if text_body:	dOpts['text'] = text_body
	if html_body:	dOpts['html'] = html_body
	if from_:		dOpts['from'] = from_
	if bcc:			dOpts['bcc'] = bcc
	if attachments:	dOpts['attachments'] = attachments

	# Call new method and return result
	return Send(to, subject, dOpts)

def Send(to, subject, opts):
	"""Send

	Sends an e-mail to one or many addresses based on a dictionary of options

	Arguments:
		to (str|str[]): The email or emails to send the email to
		subject (str): The subject of the email
		opts (dict): The options used to generate the email and any headers
						'html': str
						'text': str
						'from': str,
						'reply-to': str
						'cc': str|str[]
						'bcc': str|str[]
						'attachments': list(str|dict('body', 'filename'))
							If an attachment is a string, a local filename is
							assumed, else if we receive a dictionary, it should
							contain the filename of the file, and the raw body
							of the file
						'unsubscribe': str

	Returns:
		bool
	"""

	# Import the module vars
	global __msError, __mdSMTP

	# Init the list of total "to"s
	lTO = []

	# If from is missing
	if 'from' not in opts:
		opts['from'] = 'noreply@%s' % socket.gethostname()

	# Create a new Mime MultiPart message
	oMMP = MIMEMultipart('mixed')
	oMMP['From'] = opts['from']
	oMMP['To'] = _addresses(to)
	oMMP['Date'] = formatdate()
	oMMP['Subject'] = subject

	# Add the to
	lTO.append(to)

	# If we have a reply-to
	if 'reply-to' in opts:
		oMMP['reply-to'] = opts['reply-to']

	# If we have cc
	if 'cc' in opts:
		oMMP['Cc'] = _addresses(opts['cc'])
		lTO.append(opts['cc'])

	# If we have bcc
	if 'bcc' in opts:
		lTO.append(opts['bcc'])

	# If we have an unsubscribe string
	if 'unsubscribe' in opts:
		oMMP.add_header('List-Unsubscribe', opts['unsubscribe'])

	# Create the alternative part for the content
	oAlternative = MIMEMultipart('alternative')

	# Check that text or html body is set
	if 'text' not in opts and 'html' not in opts:
		raise ValueError('need one of "text" or "html" in Send options')

	# Attach the main message
	if 'text' in opts and opts['text']:
		oAlternative.attach(MIMEText(opts['text'], 'plain'))
	if 'html' in opts and opts['html']:
		oAlternative.attach(MIMEText(opts['html'], 'html'))

	# Add the alternative section to the email
	oMMP.attach(oAlternative)

	# If there's any attachments
	if 'attachments' in opts:

		# If we didn't get a list
		if not isinstance(opts['attachments'], (list,tuple)):
			raise ValueError('"attachments" must be a list if set')

		# Loop through the attachments
		for m in opts['attachments']:

			# If we got a string
			if isinstance(m, basestring):

				# Assume it's a file and open it
				with open(m, "rb") as rFile:
					oMMP.attach(MIMEApplication(
						rFile.read(),
						Content_Disposition='attachment; filename="%s"' % basename(m),
						Name=basename(m)
					))

			# Else if we get a dict
			elif isinstance(m, dict):

				# Add it
				oMMP.attach(MIMEApplication(
					m['body'],
					Content_Disposition='attachment; filename="%s"' % m['filename'],
					Name=m['filename']
				))

			# Unknown type
			else:
				raise ValueError('Invalid attachment', m)

	# Generate the body
	sBody = oMMP.as_string()

	# Catch any Connect or Authenticate Errors
	try:

		# Create a new instance of the SMTP class
		oSMTP = smtplib.SMTP(__mdSMTP['host'], __mdSMTP['port'])

		# If we need TLS
		if __mdSMTP['tls']:

			# Start TLS
			oSMTP.starttls()

		# If there's a username
		if __mdSMTP['user']:

			# Log in with the given credentials
			oSMTP.login(__mdSMTP['user'], __mdSMTP['passwd'])

		# Try to send the message, then close the SMTP
		oSMTP.sendmail(opts['from'], _to(lTO), sBody)
		oSMTP.close()

		# Return ok
		return OK

	# If there's a connection error
	except smtplib.SMTPConnectError as e:
		__msError = str(e.args)
		return ERROR_CONNECT

	# If there's am authentication error
	except smtplib.SMTPAuthenticationError as e:
		__msError = str(e.args)
		return ERROR_LOGIN

	# If there's any other error
	except smtplib.SMTPException as e:
		__msError = str(e.args)
		return ERROR_UNKNOWN



# coding=utf8
""" REST Module

Provides interfaces for creating and maintaining REST servers
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2018-11-11"

# Ouroboros imports
from tools import clone
import jsonb

# Python imports
import re
import sys
import traceback

# Pip imports
import bottle

# Module imports
from . import Errors, Services, Session

# Valid content types
_reContentType = re.compile(r'^application\/json; charset=utf-?8$')

# Method bytes
#	Create
C		= 0x1
CREATE	= 0x1
POST	= 0x1
#	Read
GET		= 0x2
R		= 0x2
READ	= 0x2
#	Update
PUT		= 0x4
U		= 0x4
UPDATE	= 0x4
#	Delete
D		= 0x8
DELETE	= 0x8
#	All
A		= 0xF
ALL		= 0xF
CRUD	= 0xF
#	Create, Update, Delete
CUD		= C | U | D
CREATE_UPDATE_DELETE = CUD
#	Create, Read, Delete
CRD		= C | R | D
CREATE_READ_DELETE = CRD
#	Read, Update
RU		= R | U
READ_UPDATE = RU

class _Route(object):
	"""Route

	A callable class used to store rest routes in the server
	"""

	def __init__(self, service, path, cors=None, error_callback=None, many=False):
		"""Constructor (__init__)

		Initialises an instance of the route

		Arguments:
			service (str): The service we are routing to
			path (str): The path in the service we are routing to
			environ (bool): True if the route requires request environ
			cors (dict): Optionsl CORS values
			error_callback (function): Optional callback for errors
			many (boolean): Optional, used to identify the request as a __list

		Returns:
			None
		"""
		self.service = service
		self.path = path
		self.cors = cors
		self.error_callback = error_callback
		self.many = many

	def __call__(self):
		"""Call (__call__)

		Python magic method that allows the instance to be called

		Returns:
			str
		"""

		# If CORS is enabled and the origin matches
		if self.cors and 'origin' in bottle.request.headers and self.cors.match(bottle.request.headers['origin']):
			bottle.response.headers['Access-Control-Allow-Origin'] = bottle.request.headers['origin']
			bottle.response.headers['Vary'] = 'Origin'

		# If the request is OPTIONS, set the headers and return nothing
		if bottle.request.method == 'OPTIONS':
			bottle.response.headers['Access-Control-Allow-Methods'] = 'DELETE, GET, POST, PUT, OPTIONS'
			bottle.response.headers['Access-Control-Max-Age'] = 1728000
			bottle.response.headers['Access-Control-Allow-Headers'] = 'Authorization,DNT,X-CustomHeader,Keep-Alive,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type'
			bottle.response.headers['Content-Type'] = 'text/plain charset=UTF-8'
			bottle.response.headers['Content-Length'] = 0
			bottle.request.status = 204
			return ''

		# Set the return to JSON
		bottle.response.headers['Content-Type'] = 'application/json; charset=utf-8'

		# Initialise the request details with the environment
		dReq = {
			'environment': bottle.request.environ
		}

		# If we got a Read request and the data is in the GET
		if bottle.request.method == 'GET' and 'd' in bottle.request.query:

			# Convert the GET and store the data
			try:
				dReq['data'] = jsonb.decode(bottle.request.query['d'])
			except Exception as e:
				return str(Services.Error((Errors.REST_REQUEST_DATA, '%s\n%s' % (bottle.request.query['d'], str(e)))))

		# Else we most likely got the data in the body
		else:

			# Make sure the request sent JSON
			try:
				if not _reContentType.match(bottle.request.headers['Content-Type'].lower()):
					return str(Services.Error(Errors.REST_CONTENT_TYPE))
			except KeyError:
				return str(Services.Error(Errors.REST_CONTENT_TYPE))

			# Store the data, if it's too big we need to read it rather than
			#	use getvalue
			try: sData = bottle.request.body.getvalue()
			except AttributeError as e: sData = bottle.request.body.read()

			# Make sure we have a string, not a set of bytes
			try: sData = sData.decode()
			except (UnicodeDecodeError, AttributeError): pass

			# Convert the data and store it
			try:
				if sData: dReq['data'] = jsonb.decode(sData)
			except Exception as e:
				return str(Services.Error(Errors.REST_REQUEST_DATA,'%s\n%s' % (sData, str(e))))

		# If the request should have sent a session, or one was sent anyway
		if 'Authorization' in bottle.request.headers:

			# Get the session from the Authorization token
			dReq['session'] = Session.load(bottle.request.headers['Authorization'])

			# If the session is not found
			if not dReq['session']:
				bottle.response.status = 401
				return str(Services.Error(Errors.REST_AUTHORIZATION, 'Unauthorized'))

			# Else, extend the session
			else:
				dReq['session'].extend()

		# In case the service crashes
		try:

			# Call the appropriate API method based on the HTTP/request method
			if bottle.request.method == 'DELETE':
				oResponse = Services.delete(self.service, self.path, dReq)
			elif bottle.request.method == 'POST':
				oResponse = Services.create(self.service, self.path, dReq)
			elif bottle.request.method == 'PUT':
				oResponse = Services.update(self.service, self.path, dReq)

			# If it's specifically GET
			elif bottle.request.method == 'GET':

				# If it's the special __list path
				if self.many is True:

					# If the data isn't passed or isn't an array
					if 'data' not in dReq or not isinstance(dReq['data'], list):
						oResponse = Services.Error(
							Errors.REST_REQUEST_DATA,
							'data must be an array'
						)

					# Else, if it's beyond the max
					elif len(dReq['data']) > 10:
						oResponse = Services.Error(
							Errors.REST_LIST_TO_LONG,
							'Can not request more than 10 urls via __list'
						)

					# Else, we have an array of nouns to call and return
					else:

						# Init the response
						lResponse = []

						# Go through each element in the list
						for m in dReq['data']:

							# If we got a string
							if isinstance(m, str):
								m = [m]

							# Generate unique request details for the element
							dRequest = clone(dReq)
							if len(m) == 2:
								dRequest['data'] = m[1]
							else:
								try: del dRequest['data']
								except KeyError: pass

							# Call the request and append the data to the
							#	response
							lResponse.append([
								m[0],
								Services.read(
									self.service, m[0], dRequest
								).to_dict()
							])

						# Set the response
						oResponse = Services.Response(lResponse)

				# Else, just do a normal read
				else:
					oResponse = Services.read(self.service, self.path, dReq)

		except Exception as e:
			sError = traceback.format_exc()
			print(sError, file=sys.stderr)
			if self.error_callback:
				oDetails = {
					'service': self.service,
					'method': bottle.request.method,
					'path': self.path,
					'environment': dReq['environment'],
					'traceback': sError
				}
				for s in ['data', 'session']:
					if s in dReq:
						oDetails[s] = dReq[s]
				self.error_callback(oDetails)
			oResponse = Services.Error(
				Errors.SERVICE_CRASHED,
				'%s:%s' % (self.service, self.path)
			)

		# If there's an error
		if oResponse.error_exists():

			# If it's an authorization error
			if oResponse.error['code'] == Errors.REST_AUTHORIZATION:

				# Set the http status to 401 Unauthorized
				bottle.response.status = 401

				# If the message is missing
				if oResponse.error['msg'] == '':
					oResponse.error['msg'] = 'Unauthorized'

			# Add the service and path to the call
			try: oResponse.error['service'].append([self.service, self.path])
			except KeyError: oResponse.error['service'] = [[self.service, self.path]]

		# Return the Response as a string
		return str(oResponse)

class Config(object):
	"""Config class

	Represents configuration data for connecting to/loading services
	"""

	def __contains__(self, service):
		"""__contains__

		Python magic method for checking a key exists in a dict like object

		Arguments:
			service (str): The service to check for

		Returns:
			bool
		"""
		return service in self.__services

	def __getitem__(self, service):
		"""__getitem__

		Python magic method for getting a key from a dict like object

		Arguments:
			service {str}: The service config to return

		Raises:
			KeyError

		Returns:
			mixed
		"""

		# If it's in services
		if service in self.__services:
			return self.__services[service].copy()

		# Else, throw an exception
		else:
			raise KeyError(service)

	def __init__(self, conf):
		"""Constructor

		Initialises the instance

		Arguments:
			conf {dict}: The configuration data for compiling the list of
				services and loaders

		Returns:
			Config
		"""

		# If we didn't get a dictionary for the service conf
		if not isinstance(conf, dict):
			raise ValueError('conf')

		# If we didn't get a list of services
		if 'services' not in conf:
			raise ValueError('conf.services')

		# Init the defaults if none are found
		if 'defaults' not in conf:
			conf['defaults'] = {}

		# Port values are not modified by default
		iPortMod = 0

		# If there is a port modifier
		if 'port' in conf['default']:

			# Make sure it's an integer
			try:
				iPortMod = int(conf['default']['port'])
				del conf['default']['port']
			except ValueError:
				raise ValueError('conf.default.port must be an int')

		# Initialise the list of services
		self.__services = {}

		# Loop through the list of services
		for s in conf['services']:

			# If the service doesn't point to a dict
			if not isinstance(conf['services'][s], dict):
				raise ValueError('conf.services.%s' % s)

			# Start with the default values
			dParts = conf['default'].copy()

			# Then add the service values
			dParts.update(conf['services'][s])

			# If we have no port
			if 'port' not in dParts:

				# But we have a modifier, assume we add to 80
				if iPortMod: dParts['port'] = 80 + iPortMod

			# Else add the modifier to the port passed
			else:
				dParts['port'] += iPortMod

			# Set defaults for any missing parts
			if not dParts['protocol']: dParts['protocol'] = 'http'
			if not dParts['domain']: dParts['domain'] = 'localhost'
			if 'path' not in dParts: dParts['path'] = ''
			else: dParts['path'] = '%s/' % str(dParts['path'])

			# Store the parts for the service
			self.__services[s] = dParts.copy()

			# Generate a URL from the parts and store it
			self.__services[s]['url'] = '%s://%s%s/%s' % (
				dParts['protocol'],
				dParts['domain'],
				'port' in dParts and ":%d" % dParts['port'] or '',
				dParts['path']
			)

			# If we still have no port, default to 80
			if 'port' not in self.__services[s]:
				self.__services[s]['port'] = 80

		# If there are loaders in the conf
		if 'loaders' in conf:

			# Loop through each loader
			for l in conf['loaders']:

				# If we didn't get a dict
				if not isinstance(conf['loaders'][l], dict):
					raise ValueError('conf.loaders.%s' % l)

				# Copy the parts
				dParts = conf['loaders'][l].copy()

				# If no port is set for the loader
				if 'port' not in dParts:

					# But we have a modifier, assume we add to 80
					if iPortMod: dParts['port'] = 80 + iPortMod

				# Else add the modifier to the port passed
				else:
					dParts['port'] += iPortMod

				# Store the parts for the loader
				self.__services[l] = dParts.copy()

	def __iter__(self):
		"""__iter__

		Python magic method to return an iterator for the instance

		Returns:
			iterator
		"""
		return iter(self.__services)

	def __str__(self):
		"""__str__

		Python magic method to return a string for the instance

		Returns:
			str
		"""
		return str(self.__services)

	def keys(self):
		"""services

		Returns the keys (services) in the instance

		Returns:
			str[]
		"""
		return self.__services.keys()

# Server class
class Server(bottle.Bottle):
	"""Server

	Creates an HTTP server for use with REST requests
	"""

	def __init__(self,
			routes,
			service = '',
			cors = None,
			error_callback = None,
			lists = True):
		"""Constructor (__init__)

		Instantiates the server instance

		Arguments:
			routes (dict|list): Routes to the server
			service (str): The service to use if none exists in a route
			cors (str): The regex to identify allowed domains
			error_callback (function): A function to call if any exception
										occurs
			lists (bool | dict): True to add a __list call to the default
								service, else a dictionay of uris to services

		Returns:
			None
		"""

		# Call the parent constructor first so the object is setup
		super(Server, self).__init__()

		# If cors
		if cors: cors = re.compile(cors)

		# If we have a request for a list of requests
		if lists:

			# If it's True
			if lists is True:
				lists = {'/__list': service}

			# Go through each request in the list
			for sUri in lists:

				# Add the list read route
				self.route(
					sUri,
					['OPTIONS', 'GET'],
					_Route(
						lists[sUri],	# The service to use
						sUri[1:],		# The path in the service
						cors,			# CORS regex
						error_callback,	# Callback for error
						True			# Handles many requests
					)
		)

		# If the routes are passed as a dict
		if isinstance(routes, dict):

			# Create a new list for the routes
			l = []

			# Convert them using the key as the uri
			for k,v in routes.items():
				v['uri'] = k
				l.append(v)
			routes = l

		# Go through each route
		for d in routes:

			# If there's no method passed, assume all
			if 'methods' not in d:
				lMethods = ['DELETE', 'GET', 'POST', 'PUT']

			# Else, use the bits to figure out the methods
			else:
				lMethods = ['OPTIONS']
				if d['methods'] & CREATE: lMethods.append('POST')
				if d['methods'] & DELETE: lMethods.append('DELETE')
				if d['methods'] & READ: lMethods.append('GET')
				if d['methods'] & UPDATE: lMethods.append('PUT')

			# If the service is missing, use the default
			if 'service' not in d:
				d['service'] = service

			# If the path is not passed, generate it from the uri
			if 'path' not in d:
				d['path'] = d['uri'][1:]

			# Create and add the route to the server
			self.route(
				d['uri'],
				lMethods,
				_Route(
					d['service'],	# The service to use
					d['path'],		# The path in the service
					cors,			# Optional CORS regex
					error_callback	# Optional callback for error
				)
			)

	# run method
	def run(self, server='gunicorn', host='127.0.0.1', port=8080,
			reloader=False, interval=1, quiet=False, plugins=None,
			debug=None, maxfile=20971520, **kargs):
		"""Run

		Overrides Bottle's run to default gunicorn and other fields

		Arguments:
			server (str): Server adapter to use
			host (str): Server address to bind to
			port (int): Server port to bind to
			reloader (bool): Start auto-reloading server?
			interval (int): Auto-reloader interval in seconds
			quiet (bool): Suppress output to stdout and stderr?
			plugins (list): List of plugins to the server
			debug (bool): Debug mode
			maxfile (int): Maximum size of requests

		Returns:
			None
		"""

		# Set the max file size
		bottle.BaseRequest.MEMFILE_MAX = maxfile

		# Call bottle run
		bottle.run(
			app=self, server=server, host=host, port=port, reloader=reloader,
			interval=interval, quiet=quiet, plugins=plugins, debug=debug,
			**kargs
		)



# coding=utf8
""" Email Module

Methods for sending and receiving email
"""

__author__ = "Chris Nasr"
__copyright__ = "Ouroboros Coding Inc."
__version__ = "1.0.0"
__email__ = "chris@ouroboroscoding.com"
__created__ = "2022-08-25"

# Ouroboros imports
from tools import evaluate

# Python imports
from base64 import b64decode
import platform
import re
import sys

# Module imports
from . import SMTP

last_error = ''
"""The last error generated"""

__regex = re.compile(r'^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}$')
"""E-mail address regular expression"""

__mdConf = None
"""Email conf"""

def init(conf):
	"""Init

	Initialises the module

	Returns:
		None
	"""

	# Import the module variable
	global __mdConf

	# Load email conf
	__mdConf = conf

	# If we have no hostname
	if 'hostname' not in __mdConf:
		__mdConf['hostname'] = platform.node()

	# Init the SMTP module
	SMTP.init(**__mdConf['smtp'])

def error(message):
	"""Email Error

	Send out an email with an error message

	Arguments:
		error (str): The error to email

	Returns:
		bool
	"""

	# If we have no error_to
	if 'error_to' not in __mdConf:
		raise KeyError('missing "error_to" in EMail config. Make sure to pass it in init() config.')

	# Send the email
	bRes = send({
		'to': __mdConf['error_to'],
		'subject': '%s Error' % __mdConf['hostname'],
		'text': message
	})
	if not bRes:
		print('Failed to send email: %s' % last_error, file=sys.stderr)
		return False

	# Return OK
	return True

def send(conf):
	"""Send

	Sends an email over SMTP

	Arguments:
		conf (dict): Data used to generate the email config

	Returns:
		bool
	"""

	global last_error

	# If we haven't been initialised
	if not __mdConf:
		raise RuntimeError('Module not initialised. Call .init(conf) before using')

	# Check that we have at least one type of body
	if 'html' not in conf and 'text' not in conf:
		raise ValueError('must pass one of "text" or "html"')

	# If the from is not set
	if 'from' not in conf:
		conf['from'] = 'from' in __mdConf and __mdConf['from'] or ('webmaster@%s' % __mdConf['hostname'])

	# If there's an attachment
	if 'attachments' in conf:

		# Make sure it's a list
		if not isinstance(conf['attachments'], (list,tuple)):
			conf['attachments'] = [conf['attachments']]

		# Loop through the attachments
		for i in range(len(conf['attachments'])):

			# If we didn't get a dictionary
			if not isinstance(conf['attachments'][i], dict):
				raise ValueError('attachments.%d must be a dict' % i)

			# If the fields are missing
			try:
				evaluate(conf['attachments'][i], ['body', 'filename'])
			except ValueError as e:
				raise ValueError([('attachments.%d.%s' % (i, s), 'invalid') for s in e.args])

			# Try to decode the base64
			conf['attachments'][i]['body'] = b64decode(conf['attachments'][i]['body'])

	# Send the e-mail
	iRes = SMTP.Send(
		'override' in __mdConf and __mdConf['override'] or conf['to'],
		conf['subject'],
		conf
	)

	# If there was an error
	if iRes != SMTP.OK:
		last_error = '%d %s' % (iRes, SMTP.last_error())
		return False

	# Clear the error
	last_error = ''

	# Return OK
	return True

def valid(address):
	"""Valid

	Returns true if the email address is valid

	Arguments:
		address (str): The e-mail address to verify

	Returns
		bool
	"""

	# If we get a match
	if __regex.match(address):
		return True

	# No match
	return False



""" Common Errors """

REST_REQUEST_DATA		= 100
REST_CONTENT_TYPE		= 101
REST_AUTHORIZATION		= 102
REST_LIST_TO_LONG		= 103

SERVICE_ACTION			= 200
SERVICE_STATUS			= 201
SERVICE_CONTENT_TYPE	= 202
SERVICE_UNREACHABLE		= 203
SERVICE_NOT_REGISTERED	= 204
SERVICE_NO_SUCH_NOUN	= 205
SERVICE_INTERNAL_KEY	= 206
SERVICE_CRASHED			= 207
SERVICE_NO_DATA			= 208
SERVICE_NO_SESSION		= 209